import base64
import datetime
import hashlib
import hmac
import re
import time
from os import urandom
from sys import byteorder


def make_authentication_headers(client_id: str, client_shared_secret: str, request_uri: str) -> dict:
    """ Makes a dictionary with authentication headers (used across all the services)
    using given secret keys.

    :param client_id:
    :param client_shared_secret:
    :param request_uri: URI that was used for encoding
    :return: headers dictionary
    """
    token, timestamp = encode_auth_token(client_id, client_shared_secret, request_uri)
    return {
        'Authentication': token,
        'X-Sonikpass-Authentication-Timestamp': str(timestamp),
        'X-Sonikpass-Authentication-Version': '1'
    }


def encode_auth_token(client_id: str, client_shared_secret: str, request_uri: str) -> (str, str):
    request_url = re.sub(r'.*://', '', request_uri)
    nonce = str(int.from_bytes(urandom(8), byteorder))

    key = ''.join([nonce, client_shared_secret])
    token_256 = hashlib.sha256(key.encode()).digest()
    token = token_256[:16]

    timestamp = int(time.time())
    key = ''.join([nonce, request_url, str(timestamp)])
    digest_256 = hmac.new(token, msg=key.encode(), digestmod=hashlib.sha256).digest()
    digest = digest_256[:16]

    signature = base64.b64encode(digest)

    return 'hmac {0}'.format(':'.join([client_id, nonce, signature.decode('ascii')])), timestamp


import urllib
from urllib.request import Request
from urllib.error import HTTPError
import json
import logging
from dataclasses import field
from marshmallow_dataclass import dataclass
from Activeconnect.status import Status
from Activeconnect.hmac import make_authentication_headers

logger = logging.getLogger("activeconnect")

@dataclass
class Session:
    """
        Represents an Activeconnect session.

        Attributes:

            authenticated (bool): True if the authentication request was successful (does not indicate authentication status).
            failure_reason (str): If authenticated is False indicates the reason for failure.
            session_status (Status): The status of the session - a value of Session.pending indicates authentication is in progress.
            session_token (str): Identifies the session.
            session_secret (str): Secret used to authenticate REST calls to Activeconnect service.
            status_url (str): The URL to call to get the status of the session.
            logout_url (str): The URL to call to destroy the session.

    """

    @property
    def failed(self):
        return self.session_status in [Status.failed, Status.timeout, Status.cancelled]

    @property
    def in_progress(self):
        return self.session_status in [Status.pending,Status.identifying]

    @property
    def active(self):
        return self.session_status == Status.active

    def __init__(self, session_status: Status, session_token: str, session_secret: str,
                 logout_url: str, status_url: str, failure_reason: str = None):
        self.session_status = session_status
        self.failure_reason = failure_reason
        self.session_token = session_token
        self.session_secret = session_secret
        self.status_url = status_url
        self.logout_url = logout_url

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Session):
            return  self.session_status == other.session_status and \
                    self.failure_reason == other.failure_reason and \
                    self.session_token == other.session_token and \
                    self.session_secret == other.session_secret and \
                    self.status_url == other.status_url and \
                    self.logout_url == other.logout_url
        return False

    @staticmethod
    def from_authentication_response(session_data: dict):
        """
            Creates a Session from the response to an authentication request.

            Parameters:
                session_data (dict): The data returned from a call to the Activeconnect authenticate_user API.
        """
        authentication_status = session_data.get("authentication_status")

        if authentication_status is None:
            raise ValueError("session_data does not contain the authentication_status key")

        authenticated = authentication_status.get("authenticated")

        session_status = Status[authentication_status["session_status"]] if authenticated else Status.failed
        failure_reason = None if authenticated else authentication_status["reason"]
        session_token = authentication_status["session_token"] if authenticated else None
        session_secret = authentication_status["session_secret"] if authenticated else None
        status_url = authentication_status["status_url"] if authenticated else None
        logout_url = authentication_status["logout_url"] if authenticated else None

        return Session(session_status=session_status, session_token=session_token, session_secret=session_secret,
                       logout_url=logout_url, status_url=status_url, failure_reason=failure_reason)

    session_token: str = field(metadata={"required": False, "allow_none": True})
    session_secret: str = field(metadata={"required": False, "allow_none": True})
    logout_url: str = field(metadata={"required": False, "allow_none": True})
    status_url : str = field(metadata={"required": False, "allow_none": True})
    failure_reason: str = field(metadata={"required": False, "allow_none": True})
    session_status: Status = field(metadata={"required": True})

    def get_status(self ):
        """
            Get the status of the session.

            Calls the Activeconnect service to get the status of the session.

            Returns:
            Status: The status of the session.

            """

        headers = make_authentication_headers(client_id=self.session_token,
                                                   client_shared_secret=self.session_secret,
                                                   request_uri=self.status_url)
        headers["Content-Type"] = "application/json"

        req = urllib.request.Request(self.status_url, headers=headers, method='GET')

        try:
            auth_response = urllib.request.urlopen(req)
            if 200 <= auth_response.code < 300:
                d = auth_response.read()
                j = json.loads(d)
                auth_status = j.get('authentication_status')
                self.session_status = auth_status['session_status']
                return self.session_status
        except HTTPError as ex:
            logger.warning("get status failed {}".format(ex))
        return Status.failed

    def destroy(self):
        """
            Terminates the session.

            Calls the Activeconnect service to terminate this session.

            Returns:
            Boolean: True of the session was destroyed otherwise False.

            """

        headers = make_authentication_headers(client_id=self.session_token,
                                                   client_shared_secret=self.session_secret,
                                                   request_uri=self.logout_url)
        headers["Content-Type"] = "application/json"

        req = urllib.request.Request(self.logout_url, headers=headers, method='POST')

        try:
            destroy_response = urllib.request.urlopen(req)
            if 200 <= destroy_response.code < 300:
                body = destroy_response.read()
                j = json.loads(body)

                destroyed = j.get('status')

                if destroyed is None:
                    destroyed = False

                return destroyed
        except HTTPError as ex:
            logger.warning("destroy session failed {}".format(ex))

        return False



import urllib
from urllib.request import Request
from urllib.error import HTTPError

import json
import logging

from Activeconnect.config import ACTIVE_CONNECT_SERVER
from Activeconnect.session import Session
from Activeconnect.hmac import make_authentication_headers
from Activeconnect.authentication_method import AuthenticationMethod

logger = logging.getLogger("activeconnect")


class AuthenticationAPI:
    """
        Wrapper class for the Activeconnect REST API.

        Attributes:
            application_id (str): The application id of the Activeconnect application.
            application_secret (str): The application secret of the Activeconnect application.
    """

    def __init__(self, application_id: str, application_secret: str):
        """
            The constructor for ActiveConnectApi class.

            Parameters:
                application_id (str): The application id of the Activeconnect application.
                application_secret (str): The application secret of the Activeconnect application.
        """

        self.application_id = application_id
        self.application_secret = application_secret

    def authenticate_user(self, user_id, authentication_methods=None):
        """
            Authenticates a user.

            Calls the Activeconnect service to authenticate a user.

            Parameters:
            user_id (str): The id of the user to authenticate. This is the value used when adding the user.
            authentication_methods (AuthenticationMethod[]):    The authentication methods to use. If this value is None
                                                                Activeconnect will choose the authentication methods.

            Returns:
            Session: The session information or None if the request fails.

            """

        if user_id is None:
            logger.debug("user_id is None.")
            raise ValueError("user_id cannot be None")

        encoded_user_id = urllib.parse.quote_plus(user_id)

        methods = ""

        # If we have authentication methods, check they are valid and build the query string.
        if authentication_methods is not None and len(authentication_methods) > 0:
            if not all(type(n) is AuthenticationMethod for n in authentication_methods):
                logger.debug("authentication_methods contains invalid values {}".format(authentication_methods))
                raise TypeError("authentication_methods must be an array of AuthenticationMethod or None.")

            methods = "?methods=" + ",".join(method.name for method in methods)

        url = "{}/authentication/authenticate_user/{}/{}{}".format(
            ACTIVE_CONNECT_SERVER, self.application_id, encoded_user_id, methods)

        headers = make_authentication_headers(client_id=self.application_id,
                                              client_shared_secret=self.application_secret,
                                              request_uri=url)
        headers["Content-Type"] = "application/json"

        data = json.dumps({}).encode('UTF-8')

        req = urllib.request.Request(url, headers=headers, data=data)

        try:
            auth_response = urllib.request.urlopen(req)
            if 200 <= auth_response.code < 300:
                d = auth_response.read()
                response_data = json.loads(d)
                return Session.from_authentication_response(response_data)
        except HTTPError as ex:
            logger.warn("authenticate_user failed {}".format(ex))
            return None

        return None
